//Stepper Ramp EDP by AkaPigeonKicker
//Created 3-23-2021 - James Ryan Aka Pigeon Kicker
//AkaPigeonKicker.StepperRampEDP.h
// EDP = Enable, Direction, Pulse driven stepper.
//BETA Visuino ver. 8-0-0-2 not yet availble on creation date
//Patreon supporters get PERKS and honorable mentions.
//https://www.patreon.com/AkaPigeonKicker
//Get Visuino here https://www.visuino.com
//Join my Arduino robotics and electronics MeWe.
//https://mewe.com/join/arduino_robotics_and_other_stuffs
//Join the Visuino community on MeWe.
//https://mewe.com/join/visuino-visualarduinodevelopers

#ifndef _AKAPIGEONKICKER_STEPPERRAMPEDP_h
#define _AKAPIGEONKICKER_STEPPERRAMPEDP_h

namespace AkaPigeonKicker
{
	template<
		typename T_Dir,
		typename T_Ena,
		typename T_Enabled,
		typename T_End,
		typename T_Pul,
		typename T_PullDown,
		typename T_PulseTime,
		typename T_ReturnType
	> class StepperRampEDP :
		public T_Dir,
		public T_Ena,
		public T_Enabled,
		public T_End,
		public T_Pul,
		public T_PullDown,
		public T_PulseTime,
		public T_ReturnType
		
	{
	public:
		_V_PIN_( Dir )
		_V_PIN_( Ena )
		_V_PROP_( Enabled )
		_V_PIN_( End )
		_V_PIN_( Pul )
		_V_PROP_( PullDown )
		_V_PROP_( PulseTime )
		_V_PROP_( ReturnType )

	protected:
		long TotalStep = 0;
		long TakenSteps = 1;
		long Freq = 0;
		long RunningFreq = 0;
		long RampingFreq = 0;
		long RampFreq = 0;
		long RampStep = 0;
		bool EnableState = false;
		bool NeedsUpdate = false;
		bool Moving = false;
		bool Starting = false;
		bool Stoping = false;
		bool Running = false;
		bool Reversing = true;
		long LastTime = 0;
		long CurrentTime = 0;
		long NextTime = 0;
		long FreqDiv = 0;
	
	public:
//-----
		inline void SystemStart()
		{
			RampCheck();
			DirectionChange();
		}
//-----
		inline void SystemLoopBegin()
		{
			if ( !Enabled() ) return;
			if ( !Moving ) return;
			
			CheckTime();
			CheckSteps();
			CheckUpdate();
		}
//-----
		inline void CheckUpdate()
		{
			if ( NeedsUpdate )
			{
				NeedsUpdate = false;
				TakenSteps = TakenSteps + 1;
				T_Pul::SetPinValue( PullDown() );
				delayMicroseconds( PulseTime() );
				T_Pul::SetPinValue( !PullDown() );
				if ( Running ) NextTime = (CurrentTime + RunningFreq ) ;
				if ( Starting ) NextTime = ( CurrentTime + RampFreq) ;
				if ( Stoping ) NextTime = (CurrentTime + RampFreq) ;
			}
		}
//-----
		inline void CheckTime()
		{
			CurrentTime = micros();
			if ( CurrentTime >= NextTime)
			{
				NeedsUpdate = true;
				LastTime = CurrentTime;
				if ( Starting ) RampFreq = (( RampStep -  TakenSteps ) * RampingFreq) + RunningFreq ;
				if ( Stoping ) RampFreq = (( RampStep - (TotalStep - TakenSteps)) * RampingFreq) + RunningFreq ;
			}
		}
//-----
		inline void CheckSteps()
		{
			if ( TakenSteps <= RampStep ) Starting = true;
			if ( TakenSteps > RampStep ) Starting = false;
			if ( TakenSteps <= ( RampStep + ( TotalStep - RampStep))) Running = true;
			if ( TakenSteps >= ( TotalStep - RampStep))
			{
				Running = false;
				Stoping = true;
			}
			if ( TakenSteps > TotalStep )
			{
				T_End::ClockPin();
				TakenSteps = 1;
				Starting = false;
				Running = false;
				Stoping = false;
				Moving = false;
				EnableState = !PullDown();
				T_Ena::SetPinValue( EnableState );
				if ( ReturnType() ) DirectionChange();
			}
		}
//-----
		inline void Start_o_Receive( void *_StartData )
		{
			EnableState = PullDown();
			T_Ena::SetPinValue( EnableState );
			TakenSteps = 1;
			NeedsUpdate = true;
			Moving = true;
			RampCheck();
			SystemLoopBegin();
		}
//-----
		inline void Stop_o_Receive( void *_StopData )
		{
			EnableState = !PullDown();
			T_Ena::SetPinValue( EnableState );
			Moving = false;
		}
//-----Freq = Frequency_o_Receive().GetValue();
		inline void Frequency_o_Receive( void *_FrequencyData )
		{
			Freq = *(long*)_FrequencyData;
			if ( Freq <= 0 ) Freq = 1;
		}
//-----TotalStep = TotalSteps_o_Receive().GetValue();
		inline void TotalSteps_o_Receive( void *_StepData )
		{
			TotalStep = *(long*)_StepData;
		}
//-----RampStep = RampSteps_o_Receive().GetValue();
		inline void RampSteps_o_Receive( void *_RampStepData )
		{
			RampStep = *(long*)_RampStepData;
		}
//-----
		inline void RampCheck()
		{
			if ( RampStep > TotalStep / 2 ) RampStep = TotalStep / 2;
			RampingFreq = ( 1000000 / Freq ) /  RampStep;
			if ( RampingFreq <= 0 ) RampingFreq = 1;
			RunningFreq = ( 1000000 / Freq );
		}
//-----
		inline void DirectionChange()
		{
			Reversing = !Reversing;
			T_Dir::SetPinValue( Reversing );
		}
//-----
	};
//-----
}
#endif
